#ifndef DUNE_TECTONIC_GLOBAL_RUINA_NONLINEARITY_HH
#define DUNE_TECTONIC_GLOBAL_RUINA_NONLINEARITY_HH

#include <vector>

#include <dune/common/nullptr.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include "globalnonlinearity.hh"
#include "localnonlinearity.hh"
#include "nicefunction.hh"

namespace Dune {
template <class VectorType, class MatrixType>
class GlobalRuinaNonlinearity
    : public GlobalNonlinearity<VectorType, MatrixType> {
public:
  using GlobalNonlinearity<VectorType, MatrixType>::dim;

  GlobalRuinaNonlinearity(
      shared_ptr<BlockVector<FieldVector<double, 1>> const> nodalIntegrals,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> a,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> mu,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> eta,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> normalStress,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> b,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> state,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> L, double h)
      : nodalIntegrals(nodalIntegrals),
        a(a),
        mu(mu),
        eta(eta),
        normalStress(normalStress),
        b(b),
        state(state),
        L(L),
        h(h),
        trivialNonlinearity(
            new LocalNonlinearity<dim>(make_shared<TrivialFunction const>())),
        restrictions(nodalIntegrals->size()) // TODO: can we get the size from
                                             // another place?
  {
    for (auto &x : restrictions)
      x = shared_ptr<LocalNonlinearity<dim> const>();
  }

  /*
    Return a restriction of the outer function to the i'th node.
  */
  virtual shared_ptr<LocalNonlinearity<dim> const> restriction(int i) const {
    if ((*nodalIntegrals)[i][0] == 0)
      return trivialNonlinearity;

    if (restrictions[i] != nullptr)
      return restrictions[i];

    auto const func = make_shared<RuinaFunction const>(
        (*nodalIntegrals)[i][0], (*a)[i][0], (*mu)[i][0], (*eta)[i][0],
        (*normalStress)[i][0], (*b)[i][0], (*state)[i][0], (*L)[i][0], h);
    restrictions[i] = make_shared<LocalNonlinearity<dim> const>(func);
    return restrictions[i];
  }

private:
  shared_ptr<LocalNonlinearity<dim> const> const trivialNonlinearity;

  shared_ptr<BlockVector<FieldVector<double, 1>> const> nodalIntegrals;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> a;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> mu;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> eta;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> normalStress;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> b;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> state;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> L;
  double const h;

  std::vector<shared_ptr<LocalNonlinearity<dim> const>> mutable restrictions;
};
}
#endif
