#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef srcdir
#error srcdir unset
#endif

#ifdef HAVE_IPOPT
#undef HAVE_IPOPT
#endif

#ifndef HAVE_PYTHON
#error Python is required
#endif

#include <exception>
#include <iostream>

#include <boost/format.hpp>

#include <Python.h>

#include <cmath>

#include "LambertW.h"

#include <dune/common/bitsetvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/common/stdstreams.hh>
#include <dune/common/timer.hh>
#include <dune/grid/common/mcmgmapper.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include <dune/fufem/assemblers/boundaryfunctionalassembler.hh>
#include <dune/fufem/assemblers/functionalassembler.hh>
#include <dune/fufem/assemblers/localassemblers/neumannboundaryassembler.hh>
#include <dune/fufem/assemblers/localassemblers/stvenantkirchhoffassembler.hh>
#include <dune/fufem/assemblers/localassemblers/vonmisesstressassembler.hh>
#include <dune/fufem/assemblers/operatorassembler.hh>
#include <dune/fufem/assemblers/transferoperatorassembler.hh>
#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/dunepython.hh>
#include <dune/fufem/functions/basisgridfunction.hh>
#include <dune/fufem/functions/constantfunction.hh>
#include <dune/fufem/functions/vtkbasisgridfunction.hh>
#include <dune/fufem/functionspacebases/p0basis.hh>
#include <dune/fufem/functionspacebases/p1nodalbasis.hh>
#include <dune/fufem/sharedpointermap.hh>
#include <dune/solvers/common/numproc.hh> // Solver::FULL
#include <dune/solvers/iterationsteps/blockgsstep.hh>
#include <dune/solvers/iterationsteps/multigridstep.hh>
#include <dune/solvers/iterationsteps/truncatedblockgsstep.hh>
#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/solvers/loopsolver.hh>
#include <dune/solvers/transferoperators/compressedmultigridtransfer.hh>
#include <dune/tnnmg/iterationsteps/genericnonlineargs.hh>
#include <dune/tnnmg/iterationsteps/tnnmgstep.hh>

#include <dune/tectonic/globallaursennonlinearity.hh>
#include <dune/tectonic/globalruinanonlinearity.hh>
#include <dune/tectonic/myblockproblem.hh>
#include <dune/tectonic/myconvexproblem.hh>
#include <dune/tectonic/mydirectionalconvexfunction.hh>

int const dim = 2;

template <class GridView>
void setup_boundary(GridView const &gridView,
                    Dune::BitSetVector<dim> &ignoreNodes,
                    Dune::BitSetVector<1> &neumannNodes,
                    Dune::BitSetVector<1> &frictionalNodes) {
  typedef typename GridView::template Codim<dim>::Iterator VertexLeafIterator;

  typedef Dune::MultipleCodimMultipleGeomTypeMapper<
      GridView, Dune::MCMGVertexLayout> VertexMapper;
  VertexMapper const myVertexMapper(gridView);

  size_t dirichlet_nodes = 0;
  size_t neumann_nodes = 0;
  size_t frictional_nodes = 0;
  for (VertexLeafIterator it = gridView.template begin<dim>();
       it != gridView.template end<dim>(); ++it) {
    assert(it->geometry().corners() == 1);
    Dune::FieldVector<double, dim> const coordinates = it->geometry().corner(0);
    if (coordinates[1] == 1) {
      ++dirichlet_nodes;
      size_t const id = myVertexMapper.map(*it);
      ignoreNodes[id] = true;
    } else if (coordinates[1] == 0) {
      ++frictional_nodes;
      size_t const id = myVertexMapper.map(*it);
      frictionalNodes[id] = true;
      ignoreNodes[id][1] = true; // Zero displacement in direction y
    } else if (coordinates[0] == 0 || coordinates[0] == 1) {
      ++neumann_nodes;
      size_t const id = myVertexMapper.map(*it);
      neumannNodes[id] = true;
    }
  }
  std::cout << "Number of Neumann nodes: " << neumann_nodes << std::endl;
  std::cout << "Number of Dirichlet nodes: " << dirichlet_nodes << std::endl;
  std::cout << "Number of frictional nodes: " << dirichlet_nodes << std::endl;
}

// Assembles Neumann boundary term in f
template <class GridType, class GridView, class LocalVectorType, class FEBasis>
void assemble_neumann(GridView const &gridView, FEBasis const &feBasis,
                      Dune::BitSetVector<1> const &neumannNodes,
                      Dune::BlockVector<LocalVectorType> &f,
                      Dune::VirtualFunction<double, double> const &neumann,
                      double time) { // constant sample function on neumann
                                     // boundary
  BoundaryPatch<GridView> const neumannBoundary(gridView, neumannNodes);
  LocalVectorType SampleVector(0);
  neumann.evaluate(time, SampleVector[0]);
  SampleVector[1] = 0;
  ConstantFunction<LocalVectorType, LocalVectorType> const fNeumann(
      SampleVector);
  NeumannBoundaryAssembler<GridType, LocalVectorType> neumannBoundaryAssembler(
      fNeumann);

  BoundaryFunctionalAssembler<FEBasis>(feBasis, neumannBoundary).assemble(
      neumannBoundaryAssembler, f, true); // resize and zero output vector
}

// Assembles constant 1-function on frictional boundary in nodalIntegrals
template <class GridType, class GridView, class LocalVectorType, class FEBasis>
Dune::shared_ptr<Dune::BlockVector<Dune::FieldVector<double, 1>>>
assemble_frictional(GridView const &gridView, FEBasis const &feBasis,
                    Dune::BitSetVector<1> const &frictionalNodes) {
  typedef Dune::FieldVector<double, 1> Singleton;
  BoundaryPatch<GridView> const frictionalBoundary(gridView, frictionalNodes);
  ConstantFunction<LocalVectorType, Singleton> const constantOneFunction(1);
  NeumannBoundaryAssembler<GridType, Singleton> frictionalBoundaryAssembler(
      constantOneFunction);

  auto const nodalIntegrals = Dune::make_shared<Dune::BlockVector<Singleton>>();
  BoundaryFunctionalAssembler<FEBasis>(feBasis, frictionalBoundary)
      .assemble(frictionalBoundaryAssembler, *nodalIntegrals,
                true); // resize and zero output vector
  return nodalIntegrals;
}

template <class VectorType, class MatrixType>
Dune::shared_ptr<Dune::GlobalNonlinearity<VectorType, MatrixType> const>
assemble_nonlinearity(
    int size, Dune::ParameterTree const &parset,
    Dune::shared_ptr<Dune::BlockVector<Dune::FieldVector<double, 1>>>
        nodalIntegrals,
    Dune::shared_ptr<Dune::BlockVector<Dune::FieldVector<double, 1>>> state,
    double h) {
  typedef Dune::BlockVector<Dune::FieldVector<double, 1>> SingletonVectorType;
  // {{{ Assemble terms for the nonlinearity
  auto mu = Dune::make_shared<SingletonVectorType>(size);
  *mu = parset.get<double>("boundary.friction.mu");

  auto normalStress = Dune::make_shared<SingletonVectorType>(size);
  *normalStress = parset.get<double>("boundary.friction.normalstress");

  std::string const friction_model =
      parset.get<std::string>("boundary.friction.model");
  if (friction_model == std::string("Ruina")) {
    auto a = Dune::make_shared<SingletonVectorType>(size);
    *a = parset.get<double>("boundary.friction.ruina.a");

    auto eta = Dune::make_shared<SingletonVectorType>(size);
    *eta = parset.get<double>("boundary.friction.eta");

    auto b = Dune::make_shared<SingletonVectorType>(size);
    *b = parset.get<double>("boundary.friction.ruina.b");

    auto L = Dune::make_shared<SingletonVectorType>(size);
    *L = parset.get<double>("boundary.friction.ruina.L");

    return Dune::make_shared<
        Dune::GlobalRuinaNonlinearity<VectorType, MatrixType> const>(
        nodalIntegrals, a, mu, eta, normalStress, b, state, L, h);
  } else if (friction_model == std::string("Laursen")) {
    return
        // TODO: take state and h into account
        Dune::make_shared<
            Dune::GlobalLaursenNonlinearity<Dune::LinearFunction, VectorType,
                                            MatrixType> const>(mu, normalStress,
                                                               nodalIntegrals);
  } else {
    assert(false);
  }
}

// The nonlinearity exp(-x)
class DecayingExponential {
public:
  typedef Dune::FieldVector<double, 1> VectorType;
  typedef Dune::FieldMatrix<double, 1, 1> MatrixType;

  double operator()(VectorType const &u) const { return exp(-u[0]); }

  void directionalSubDiff(VectorType const &u, VectorType const &v,
                          Interval<double> &D) const {
    D[0] = D[1] = v[0] * (-exp(-u[0]));
  }

  void directionalDomain(VectorType const &, VectorType const &,
                         Interval<double> &dom) const {
    dom[0] = -std::numeric_limits<double>::max();
    dom[1] = std::numeric_limits<double>::max();
  }
};

double compute_state_update_bisection(double h, double unorm, double L,
                                      double old_state) {
  MyDirectionalConvexFunction<DecayingExponential> const J(
      1.0 / h, (old_state - unorm / L) / h, DecayingExponential(), 0, 1);
  int bisectionsteps = 0;
  Bisection const bisection(0.0, 1.0, 1e-12, true, 0);    // TODO
  return bisection.minimize(J, 0.0, 0.0, bisectionsteps); // TODO
}

double compute_state_update_lambert(double h, double unorm, double L,
                                    double old_state) {
  double const rhs = unorm / L - old_state;
  return LambertW(0, h * exp(rhs)) - rhs;
}

int main(int argc, char *argv[]) {
  try {
    typedef SharedPointerMap<std::string, Dune::VirtualFunction<double, double>>
    FunctionMap;
    FunctionMap functions;
    {
      Python::start();

      Python::run("import sys");
      Python::run("sys.path.append('" srcdir "')");

      Python::import("one-body-sample_neumann")
          .get("Functions")
          .toC<FunctionMap::Base>(functions);
    }

    Dune::ParameterTree parset;
    Dune::ParameterTreeParser::readINITree(srcdir "/one-body-sample.parset",
                                           parset); // FIXME
    Dune::ParameterTreeParser::readOptions(argc, argv, parset);

    Dune::Timer timer;

    typedef Dune::FieldVector<double, dim> SmallVector;
    typedef Dune::FieldMatrix<double, dim, dim> SmallMatrix;
    typedef Dune::BCRSMatrix<SmallMatrix> OperatorType;
    typedef Dune::BlockVector<SmallVector> VectorType;
    typedef Dune::BlockVector<Dune::FieldVector<double, 1>> SingletonVectorType;

    auto const E = parset.get<double>("body.E");
    auto const nu = parset.get<double>("body.nu");
    auto const solver_tolerance = parset.get<double>("solver.tolerance");

    auto const refinements = parset.get<size_t>("grid.refinements");
    size_t const levels = refinements + 1;

    auto const verbose = parset.get<bool>("verbose");
    Solver::VerbosityMode const verbosity =
        verbose ? Solver::FULL : Solver::QUIET;

    // {{{ Set up grid
    typedef Dune::YaspGrid<dim> GridType;
    Dune::FieldVector<double, dim> const end_points(
        1); // nth dimension (zero-indexed) goes from 0 to end_points[n]
    auto grid = Dune::make_shared<GridType>(
        end_points,
        Dune::FieldVector<int, dim>(1), // number of elements in each direction
        Dune::FieldVector<bool, dim>(false), // non-periodic in each direction
        0);                                  // zero overlap (whatever that is)
    grid->globalRefine(refinements);

    typedef GridType::LeafGridView GridView;
    GridView const leafView = grid->leafView();
    // }}}

    // Set up bases
    typedef P0Basis<GridView, double> P0Basis;
    typedef P1NodalBasis<GridView, double> P1Basis;
    P0Basis const p0Basis(leafView);
    P1Basis const p1Basis(leafView);

    // Assemble elastic force on the body
    StVenantKirchhoffAssembler<GridType, P1Basis::LocalFiniteElement,
                               P1Basis::LocalFiniteElement> const
    localStiffness(E, nu);
    OperatorType stiffnessMatrix;
    timer.reset();
    OperatorAssembler<P1Basis, P1Basis>(p1Basis, p1Basis)
        .assemble(localStiffness, stiffnessMatrix);
    std::cout << "Assembled stiffness matrix in " << timer.elapsed() << "s"
              << std::endl;
    EnergyNorm<OperatorType, VectorType> energyNorm(stiffnessMatrix);

    // Set up the boundary
    Dune::BitSetVector<dim> ignoreNodes(grid->size(grid->maxLevel(), dim),
                                        false);
    Dune::BitSetVector<1> neumannNodes(grid->size(grid->maxLevel(), dim),
                                       false);
    Dune::BitSetVector<1> frictionalNodes(grid->size(grid->maxLevel(), dim),
                                          false);
    setup_boundary(leafView, ignoreNodes, neumannNodes, frictionalNodes);

    typedef MyConvexProblem<OperatorType, VectorType> MyConvexProblemType;
    typedef MyBlockProblem<MyConvexProblemType> MyBlockProblemType;
    GenericNonlinearGS<MyBlockProblemType> nonlinearGSStep;
    nonlinearGSStep.ignoreNodes_ = &ignoreNodes;

    VectorType u1(grid->size(grid->maxLevel(), dim));
    u1 = 0.0; // Has to be zero!
    VectorType u2 = u1;
    VectorType u3 = u1;
    VectorType u4 = u1;
    VectorType u5 = u1;

    SingletonVectorType s4_old(grid->size(grid->maxLevel(), dim));
    SingletonVectorType s5_old(grid->size(grid->maxLevel(), dim));
    s4_old = parset.get<double>("boundary.friction.state.initial");
    s5_old = s4_old;

    VectorType u1_diff(grid->size(grid->maxLevel(), dim));
    u1_diff = 0.0; // Has to be zero!
    VectorType u2_diff = u1_diff;
    VectorType u3_diff = u1_diff;
    VectorType u4_diff = u1_diff;
    VectorType u5_diff = u1_diff;

    auto s4_new = Dune::make_shared<SingletonVectorType>(
        grid->size(grid->maxLevel(), dim));
    auto s5_new = Dune::make_shared<SingletonVectorType>(
        grid->size(grid->maxLevel(), dim));
    *s4_new = s4_old;
    *s5_new = *s4_new;

    SingletonVectorType vonMisesStress;

    VectorType b1;
    VectorType b2;
    VectorType b3;
    VectorType b4;
    VectorType b5;

    auto const nodalIntegrals =
        assemble_frictional<GridType, GridView, SmallVector, P1Basis>(
            leafView, p1Basis, frictionalNodes);
    // {{{ Set up TNNMG solver
    // linear iteration step components
    TruncatedBlockGSStep<OperatorType, VectorType> linearBaseSolverStep;
    EnergyNorm<OperatorType, VectorType> baseEnergyNorm(linearBaseSolverStep);
    LoopSolver<VectorType> linearBaseSolver(
        &linearBaseSolverStep,
        parset.get<int>("solver.tnnmg.linear.maxiterations"), solver_tolerance,
        &baseEnergyNorm, Solver::QUIET);
    TruncatedBlockGSStep<OperatorType, VectorType> linearPresmoother;
    TruncatedBlockGSStep<OperatorType, VectorType> linearPostsmoother;

    // linear iteration step
    MultigridStep<OperatorType, VectorType> linearIterationStep;
    linearIterationStep.setNumberOfLevels(levels);
    linearIterationStep.setMGType(parset.get<int>("solver.tnnmg.linear.mu"),
                                  parset.get<int>("solver.tnnmg.linear.nu1"),
                                  parset.get<int>("solver.tnnmg.linear.nu2"));
    linearIterationStep.basesolver_ = &linearBaseSolver;
    linearIterationStep.setSmoother(&linearPresmoother, &linearPostsmoother);

    // transfer operators
    std::vector<CompressedMultigridTransfer<VectorType> *> transferOperators(
        refinements);
    for (auto &x : transferOperators)
      x = new CompressedMultigridTransfer<VectorType>;
    TransferOperatorAssembler<GridType>(*grid)
        .assembleOperatorPointerHierarchy(transferOperators);
    linearIterationStep.setTransferOperators(transferOperators);

    // tnnmg iteration step
    typedef GenericNonlinearGS<MyBlockProblemType> NonlinearSmootherType;
    NonlinearSmootherType nonlinearSmoother;
    TruncatedNonsmoothNewtonMultigrid<MyBlockProblemType, NonlinearSmootherType>
    multigridStep(linearIterationStep, nonlinearSmoother);
    multigridStep.setSmoothingSteps(parset.get<int>("solver.tnnmg.main.nu1"),
                                    parset.get<int>("solver.tnnmg.main.mu"),
                                    parset.get<int>("solver.tnnmg.main.nu2"));
    multigridStep.ignoreNodes_ = &ignoreNodes;
    // }}}

    std::fstream octave_writer("data", std::fstream::out);
    timer.reset();
    auto const timesteps = parset.get<size_t>("timesteps");
    octave_writer << "# name: A" << std::endl << "# type: matrix" << std::endl
                  << "# rows: " << timesteps << std::endl << "# columns: 3"
                  << std::endl;
    double const h = 1.0 / timesteps;
    for (size_t run = 1; run <= timesteps; ++run) {
      if (parset.get<bool>("printProgress")) {
        std::cout << ".";
        std::cout.flush();
      }

      // b = neumann
      assemble_neumann<GridType, GridView, SmallVector, P1Basis>(
          leafView, p1Basis, neumannNodes, b1, functions.get("sampleFunction"),
          h * run);
      b2 = b1;
      b3 = b1;
      b4 = b1;
      b5 = b1;

      // b -= linear update
      stiffnessMatrix.mmv(u1, b1);
      stiffnessMatrix.mmv(u2, b2);
      stiffnessMatrix.mmv(u3, b3);
      stiffnessMatrix.mmv(u4, b4);
      stiffnessMatrix.mmv(u5, b5);

      if (parset.get<bool>("solver.nonlineargs.use")) {
        auto state =
            Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>(
                grid->size(grid->maxLevel(), dim));
        *state = 0.0;
        auto myGlobalNonlinearity =
            assemble_nonlinearity<VectorType, OperatorType>(
                grid->size(grid->maxLevel(), dim), parset, nodalIntegrals,
                state, h);
        MyConvexProblemType const myConvexProblem(stiffnessMatrix,
                                                  *myGlobalNonlinearity, b1);
        MyBlockProblemType myBlockProblem(parset, myConvexProblem);
        nonlinearGSStep.setProblem(u1_diff, myBlockProblem);

        LoopSolver<VectorType> solver(
            &nonlinearGSStep,
            parset.get<size_t>("solver.nonlineargs.maxiterations"),
            solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u1 += u1_diff;

      if (parset.get<bool>("solver.tnnmg.use")) {
        for (int state_fpi = 0;
             state_fpi < parset.get<int>("solver.tnnmg.fixed_point_iterations");
             ++state_fpi) {
          auto myGlobalNonlinearity =
              assemble_nonlinearity<VectorType, OperatorType>(
                  grid->size(grid->maxLevel(), dim), parset, nodalIntegrals,
                  s4_new, h);
          MyConvexProblemType const myConvexProblem(stiffnessMatrix,
                                                    *myGlobalNonlinearity, b4);

          MyBlockProblemType myBlockProblem(parset, myConvexProblem);
          multigridStep.setProblem(u4_diff, myBlockProblem);

          LoopSolver<VectorType> overallSolver(
              &multigridStep, parset.get<size_t>("solver.tnnmg.maxiterations"),
              solver_tolerance, &energyNorm, verbosity);
          overallSolver.solve();

          if (!parset.get<bool>("boundary.friction.state.evolve"))
            continue;

          for (size_t i = 0; i < frictionalNodes.size(); ++i) {
            if (frictionalNodes[i][0]) {
              double const L = parset.get<double>("boundary.friction.ruina.L");
              double const unorm = u4_diff[i].two_norm();

              double ret1 =
                  compute_state_update_bisection(h, unorm, L, s4_old[i][0]);
              assert(std::abs(1.0 / h * ret1 - (s4_old[i] - unorm / L) / h -
                              exp(-ret1)) < 1e-11);

              double ret2 =
                  compute_state_update_lambert(h, unorm, L, s4_old[i][0]);
              assert(std::abs(1.0 / h * ret2 - (s4_old[i] - unorm / L) / h -
                              exp(-ret2)) < 1e-11);

              assert(std::abs(ret2 - ret1) < 1e-14);

              (*s4_new)[i][0] = ret1;
            }
          }
        }

        if (parset.get<bool>("printEvolution")) {
          // Find the first node that belongs to the frictional boundary
          for (size_t i = 0; i < frictionalNodes.size(); ++i) {
            if (!frictionalNodes[i][0])
              continue;

            boost::format const formatter("s[%03d] = %+3e, "
                                          "%|40t|u[%03d] = %+3e");
            std::cout << boost::format(formatter) % run % (*s4_new)[i] % run %
                             u4[i] << std::endl;
            double out;
            functions.get("sampleFunction").evaluate(h * run, out);
            octave_writer << (*s4_new)[i] << " " << u4[i][0] * 1e6 << " " << out
                          << std::endl;
            break;
          }
        }
      }

      u4 += u4_diff;
      s4_old = *s4_new;

      if (parset.get<bool>("benchmarks.fpi.enable")) {
        for (int state_fpi = 0;
             state_fpi < parset.get<int>("benchmarks.fpi.iterations");
             ++state_fpi) {
          auto myGlobalNonlinearity =
              assemble_nonlinearity<VectorType, OperatorType>(
                  grid->size(grid->maxLevel(), dim), parset, nodalIntegrals,
                  s5_new, h);
          MyConvexProblemType const myConvexProblem(stiffnessMatrix,
                                                    *myGlobalNonlinearity, b5);

          MyBlockProblemType myBlockProblem(parset, myConvexProblem);
          multigridStep.setProblem(u5_diff, myBlockProblem);

          LoopSolver<VectorType> overallSolver(
              &multigridStep, parset.get<size_t>("solver.tnnmg.maxiterations"),
              solver_tolerance, &energyNorm, verbosity);
          overallSolver.solve();

          if (!parset.get<bool>("boundary.friction.state.evolve"))
            continue;

          for (size_t i = 0; i < frictionalNodes.size(); ++i) {
            if (frictionalNodes[i][0]) {
              double const L = parset.get<double>("boundary.friction.ruina.L");
              double const unorm = u5_diff[i].two_norm();

              double ret1 =
                  compute_state_update_bisection(h, unorm, L, s5_old[i][0]);
              assert(std::abs(1.0 / h * ret1 - (s5_old[i] - unorm / L) / h -
                              exp(-ret1)) < 1e-12);
              double ret2 =
                  compute_state_update_lambert(h, unorm, L, s5_old[i][0]);
              assert(std::abs(1.0 / h * ret2 - (s5_old[i] - unorm / L) / h -
                              exp(-ret2)) < 1e-12);
              assert(std::abs(ret2 - ret1) < 1e-14);

              (*s5_new)[i][0] = ret1;
            }
          }
        }
      }

      u5 += u5_diff;
      s5_old = *s5_new;

      if (parset.get<bool>("benchmarks.fpi.enable"))
        std::cout << energyNorm.diff(u5, u4) / energyNorm(u5) << std::endl;

      // Compute von Mises stress and write everything to a file
      if (parset.get<bool>("writeVTK")) {
        auto const displacement =
            Dune::make_shared<BasisGridFunction<P1Basis, VectorType> const>(
                p1Basis, u4);
        VonMisesStressAssembler<GridType> localStressAssembler(E, nu,
                                                               displacement);
        FunctionalAssembler<P0Basis>(p0Basis)
            .assemble(localStressAssembler, vonMisesStress, true);

        auto const displacement_ptr =
            Dune::make_shared<VTKBasisGridFunction<P1Basis, VectorType> const>(
                p1Basis, u4, "displacement");
        auto const state_ptr = Dune::make_shared<
            VTKBasisGridFunction<P1Basis, SingletonVectorType> const>(
            p1Basis, *s4_new, "state");
        auto const vonmises_ptr = Dune::make_shared<
            VTKBasisGridFunction<P0Basis, SingletonVectorType> const>(
            p0Basis, vonMisesStress, "stress");
        Dune::VTKWriter<GridView> writer(leafView);
        writer.addVertexData(state_ptr);
        writer.addVertexData(displacement_ptr);
        writer.addCellData(vonmises_ptr);

        std::string const filename((boost::format("obs%d") % run).str());
        writer.write(filename.c_str());
      }

      if (parset.get<bool>("solver.gs.use")) {
        BlockGSStep<OperatorType, VectorType> blockGSStep(stiffnessMatrix,
                                                          u2_diff, b2);
        blockGSStep.ignoreNodes_ = &ignoreNodes;

        LoopSolver<VectorType> solver(
            &blockGSStep, parset.get<size_t>("solver.gs.maxiterations"),
            solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u2 += u2_diff;

      // Use a linear solver for comparison; should return roughly the
      // same results if phi vanishes (e.g. because the normalstress is zero)
      if (parset.get<bool>("solver.truncatedgs.use")) {
        TruncatedBlockGSStep<OperatorType, VectorType> blockGSStep(
            stiffnessMatrix, u3_diff, b3);
        blockGSStep.ignoreNodes_ = &ignoreNodes;

        LoopSolver<VectorType> solver(
            &blockGSStep,
            parset.get<size_t>("solver.truncatedgs.maxiterations"),
            solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u3 += u3_diff;
    }
    std::cout << std::endl;
    std::cout << "Making " << timesteps << " time steps took "
              << timer.elapsed() << "s" << std::endl;

    // Clean up after the TNNMG solver
    for (auto &x : transferOperators)
      delete x;

    octave_writer.close();

    if (parset.get<bool>("printFrictionalBoundary")) {
      // Print displacement on frictional boundary
      boost::format const formatter("u1[%02d] = %+3e, "
                                    "%|40t|u2[%02d] = %+3e, "
                                    "%|80t|u3[%02d] = %+3e, "
                                    "%|120t|u4[%02d] = %+3e");
      for (size_t i = 0; i < frictionalNodes.size(); ++i)
        if (frictionalNodes[i][0])
          std::cout << boost::format(formatter) % i % u1[i] % i % u2[i] % i %
                           u3[i] % i % u4[i] << std::endl;
    }
    Python::stop();
  }
  catch (Dune::Exception &e) {
    Dune::derr << "Dune reported error: " << e << std::endl;
  }
  catch (std::exception &e) {
    std::cerr << "Standard exception: " << e.what() << std::endl;
  }
}
